#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <assert.h>
#include <math.h>
#include "clock.h"
#include "gtree.h"

#ifndef TSIZE
#define TSIZE 4*1024
#endif
#ifndef REPEAT
#define REPEAT 1000
#endif

/* Lookup for a node with the 'key' identifier
   in the tree 'gt1'. The traversal of the tree
   is performed thanks to a breathfirst algorithm.
*/
inline gtree_linked* gtree_linked_lookup(gtree_linked *gt1, 
					 int key, queue_gt1 *queue1) {
  register gtree_linked *node, *child;
  
  if (gt1 == NULL) return NULL;
  INITQUEUE(queue1);
  ENQUEUE(queue1,gt1);
  while (NOTEMPTYQUEUE(queue1)) {
    DEQUEUE(queue1,node);
    if ((node->object).key == key) return node;
    child = node->leftChild;
    while (child != NULL) {
      ENQUEUE(queue1,child);
      child = child->rightSibling;
    }
  }
  return NULL;
}

/* Allocate a single node for tree of 
   data types gtree_linked */
gtree_linked *gtree_linked_malloc() {
  return (gtree_linked *) malloc(sizeof(gtree_linked));
}

/* Append a node filled with 'obj' data to the
   tree 'gt1'. The new node is added to the
   children list of the node that is characterized
   by 'parent_key'.
   Variable 'queue1' is a temporary queue used
   to perform the traversal of the tree. 
*/
inline void gtree_linked_append (gtree_linked *gt1, data *obj, 
				 int parent_key, queue_gt1 *queue1) {
  gtree_linked *node;
  gtree_linked *parent;
  node = gtree_linked_malloc();
  assert (node!=NULL);
  node->object       = *obj;
  node->leftChild    = NULL;
  node->rightSibling = NULL;
  /* Lookup for the direct 'parent' of 'node',
     in order to add 'node' as a child of
     this 'parent'. */
  parent = gtree_linked_lookup(gt1, parent_key, queue1);
  assert (parent!=NULL);
  if (parent != node) {
    /* Exclude the Root node case,
       for which we do not need to 
       add the node to the list of children
       of any parent */
    if (parent->leftChild == NULL) {
      parent->leftChild = node;
    } else {
      parent = parent->leftChild;
      while (parent->rightSibling != NULL) 
	parent = parent->rightSibling;
      parent->rightSibling = node;
    }
  } 
}

/* Allocate and fill a tree 'gt1' with 'size'
   nodes. The 'key' and 'value' encompassed within
   each node are generated automatically.
*/
void alloc_fill_gtree_linked(gtree_linked *gt1, int size, 
			     queue_gt1 *queue1) {
  struct sdata  dat;
  register int           parent_key;
  register long long int rand_num, i;
  
  /* assert((size&(~size+1))==size); */
  /* Initialize the random generator */
  rand_num = MYSEED;
  /* Initialise the root node data */
  dat.val = 2.;
  dat.key = 0;
  /* Add the root node in gt1 */
  gt1->rightSibling = NULL;
  gt1->leftChild    = NULL;
  gt1->object       = dat;

  /* Add all nodes in the tree 'gt1' */
  for (i=1; i<size; i++) {
    /* Congruantial random number generator 
       used to choose randomly a parent of the 
       node. */
    MYRAND(rand_num);
    parent_key = 4*(rand_num % i);
    /* Congruantial random number generator used 
       to choose randomly the floating point value 
       stored in the node */
    MYRAND(rand_num);
    dat.val = IRANDMAX*rand_num;
    dat.key = 4*i;
    gtree_linked_append(gt1, &dat, parent_key, queue1);
  }
}

/* Lookup for a node with the 'key' identifier
   in the tree 'gt3'. The traversal of the tree
   is simplified and accelerated by the fact 
   that all nodes are in a single array 
   'gt3->array'.
*/
inline int gtree_array_lookup(gtree_array *gt3, int key) {
  register gtree_array_elt *node;
  register int i;
  for (i=0; i<gt3->size; i++) {
    node = &(gt3->array[i]);
    if ((node->object).key == key) 
      return i;
  }
  return -1;
}

/* Append a node filled with 'obj' data to the
   tree 'pgt3'. The new node is added to the
   children list of the node that is characterized
   by 'parent_key'.
*/
inline void gtree_array_append (gtree_array *pgt3, data *obj, 
			   int parent_key) {
  gtree_array_elt *node;
  /* If the dynamic array has reached the
     maximum of its capacity, resizes it
     and doubles its capacity 
  */
  if (pgt3->size == pgt3->capacity) {
    element *newarray;
    int newcapacity;
    if (pgt3->capacity == 0) {
      newcapacity=1;
      newarray = valloc(newcapacity*sizeof(element));
    } else {
      newcapacity = 2 * pgt3->capacity;
      newarray = (element*)memalign(4096,newcapacity*sizeof(element));
      memcpy(newarray,pgt3->array,pgt3->capacity*sizeof(element));
      free(pgt3->array);
    }
    pgt3->capacity = newcapacity;
    pgt3->array    = newarray;
  }   
  /* Take the first available node of the dynamic 
     array and fill it with data coming from 
     parameter 'obj'. 
  */
  node = &(pgt3->array[pgt3->size]);
  node->object = *obj;
  node->leftChild = -1;
  node->rightSibling = -1;
  node->parent = -1;
  /* Exclude the Root node case (if parent_key < 0)
     where we do not need to find a parent node.
  */
  if (parent_key >= 0) {
    gtree_array_elt *aux;
    int         aux_index;
    /* Lookup for a node that is characterized 
       by 'parent_key'.
    */
    aux_index = gtree_array_lookup(pgt3, parent_key);
    assert(aux_index != -1);
    aux = &(pgt3->array[aux_index]);
    node->parent = aux_index;
    /* Go to the end of the children list
       of the parent to add the new node.
    */
    if (aux->leftChild < 0) {
      aux->leftChild = pgt3->size;
    } else {
      aux = &(pgt3->array[aux->leftChild]);
      while (aux->rightSibling >= 0) 
	aux = &(pgt3->array[aux->rightSibling]);
      aux->rightSibling = pgt3->size;
    }
  } 
  (pgt3->size)++;
}


/* Allocate and fill a tree 'gt3' with 'size'
   nodes. The 'key' and 'value' encompassed 
   within each node are generated automatically.
*/
void alloc_fill_gtree_array(gtree_array *gt3, int size) {
  int     parent_key;
  struct sdata dat;
  register long long int rand_num,i;

  /*assert((size&(~size+1))==size);*/
  /* Initialize the random generator */
  rand_num = MYSEED;
  /* Initialise the root node data */
  dat.val = 2.;
  dat.key = 0;
  /* Add the root node in gt3 */
  gtree_array_append(gt3, &dat, -1);
  
  /* Add all nodes in the tree 'gt3' */
  for (i=1; i<size; i++) {
    /* Congruantial random number generator 
       used to choose randomly a parent of the 
       node. */
    MYRAND(rand_num);
    parent_key = 4*(rand_num % i);
    /* Congruantial random number generator used 
       to choose randomly the floating point value 
       stored in the node */
    MYRAND(rand_num);
    dat.val = IRANDMAX*rand_num;
    dat.key = 4*i;
    gtree_array_append(gt3, &dat, parent_key);
  }
}

/* Print nodes of the tree 'gt3'. 
   The traversal of the tree
   is simplified and accelerated by the fact 
   that all nodes are in a single array 
   'pgt3->array'.
*/
void gtree_array_traversal(gtree_array *pgt3) {
  register int     i;
  register gtree_array_elt *node;
  
  for (i=0; i<pgt3->size; i++) {
    node = &(pgt3->array[i]);
    printf("node key %10ld  parent %10d leftChild %10d rightSibling %10d\n",
	   (long int)(node->object).key, node->parent, node->leftChild, node->rightSibling);
  }
}

/* Compute the sum of all values stored in 
   the nodes of the tree 'gt3'. 
   The traversal of the tree
   is performed in scanning the array
   of nodes 'pgt3->array'.
*/
double gtree_array_sumval(gtree_array *gt3) {
  register int     i;
  register double  sum1 = 0, sum2 = 0;
  register int sdiv2 = gt3->size/2;

  // if 2 divides gt3 size, optimize
  // the memory access
  if (sdiv2*2!=gt3->size) {
    for (i=0; i<gt3->size; i++) {
      sum1 += gt3->array[i].object.val;
    }
  } else {
    for (i=0; i<sdiv2; i++) {
      sum1 += gt3->array[i].object.val;
      sum2 += gt3->array[sdiv2+i].object.val;
    }
  }
  return(sum1+sum2);
}

/* Compute the sum of all values stored in 
   the nodes of the tree 'gt1'. 
   The traversal of the tree
   is performed with a breadthfirst algorithm
   of 'gt1'.
*/ 
double gtree_linked_breadthfirst_sumval(gtree_linked *gt1, 
					queue_gt1 *queue1) {
  register double  sum;
  register gtree_linked *node, *child;
  
  /* If the tree is void, return immediatly */
  if (gt1 == NULL) return 0.;
  INITQUEUE(queue1);
  sum = 0;
  /* Add the root node in the queue */
  ENQUEUE(queue1,gt1);
  while (NOTEMPTYQUEUE(queue1)) {
    /* dequeue the node at the back of the queue */
    DEQUEUE(queue1,node);
    sum += (node->object).val; 
    child = node->leftChild;
    while (child != NULL) {
      /* add children at the front of the queue */
      ENQUEUE(queue1,child);
      child = child->rightSibling;
    }
  }
  return sum;
}


/* Compute the sum of all values stored in 
   the nodes of the tree 'gt3'. 
   The traversal of the tree
   is performed with a breadthfirst algorithm
   of 'gt3'.
*/ 
double gtree_array_breadthfirst_sumval(gtree_array *gt3, 
				  queue_gt3 *queue3) {
  register double   sum;
  register gtree_array_elt *node, *pchild;
  register int32_t child;
  if (gt3->size == 0) return 0;
  INITQUEUE(queue3);
  sum = 0;
  /* Add the root node in the queue */
  ENQUEUE(queue3,&(gt3->array[0]));
  while (NOTEMPTYQUEUE(queue3)) {
    /* dequeue the node at the back of the queue */
    DEQUEUE(queue3,node);
    sum += (node->object).val; 
    child = node->leftChild;
    while (child >= 0 ) {
      pchild = &(gt3->array[child]);
      /* add children at the front of the queue */
      ENQUEUE(queue3,pchild);
      child = pchild->rightSibling;
    }
  }
  return sum;
}

/* Print all nodes stored in the tree 'gt3'. 
   The traversal of the tree
   is performed with a breadthfirst algorithm
   of 'gt3'.
*/ 
void gtree_array_breadthfirst_traversal(gtree_array *gt3, 
				   queue_gt3 *queue3) {
  register gtree_array_elt *node, *pchild;
  register int32_t child, endmark, height;
  if (gt3->size == 0) return;
  INITQUEUE(queue3);
  ENQUEUE(queue3,&(gt3->array[0])); 
  endmark = queue3->end;
  height = 0;
  while (NOTEMPTYQUEUE(queue3)) {
    /* extract node from queue: 
       (nodes of height=n) */
    DEQUEUE(queue3,node);
    printf("h %6d node key %10ld  parent %10d leftChild %10d rightSibling %10d\n",
	   height,(long int)(node->object).key, node->parent, node->leftChild, node->rightSibling);
    /* list the children node and add it to queue 
       (nodes of height=n+1) */
    child = node->leftChild;
    while (child >= 0 ) {
      pchild = &(gt3->array[child]);
      ENQUEUE(queue3,pchild);
      child = pchild->rightSibling;
    }
    /* if we reach the end of nodes list
       at this height, increment variable 'height' */
    if (endmark == queue3->start) {
      endmark = queue3->end;
      height++;
    }
  }
}      

/* Copy all nodes stored in the tree 'gt3' in
   the dynamic array 'copy'.
   The traversal of the tree
   is performed with a breadthfirst algorithm
   of 'gt3'.
*/ 
void gtree_array_breadthfirst_copy(gtree_array *gt3, gtree_array *copy, 
			      queue_gt3 *queue3) {
  register gtree_array_elt *node, *pchild;
  register int32_t child, parent_key;
  if (gt3->size == 0) return;
  /* reset the dynamic array 'copy' */
  copy->size = 0;
  INITQUEUE(queue3);
  ENQUEUE(queue3,&(gt3->array[0]));
  while (NOTEMPTYQUEUE(queue3)) {
    /* extract node from queue: 
       (nodes of height=n) */
    DEQUEUE(queue3,node);
    /* Add the actual node to the 'copy' data */
    if (node->parent < 0) {
      gtree_array_append(copy, &(node->object), -1);
    } else {
      parent_key = (gt3->array[node->parent]).object.key;
      gtree_array_append(copy, &(node->object), parent_key);
    }
    /* list the children node and add it to queue 
       (nodes of height=n+1) */
    child = node->leftChild;
    while (child >= 0 ) {
      pchild = &(gt3->array[child]);
      ENQUEUE(queue3,pchild);
      child = pchild->rightSibling;
    }
  }
}      

/* Perform several times the call the subroutine
   'gtree_array_sumval'. The number of times is parameteried
   by 'repeat' parameter.
*/ 
double repeat_sum_gtree_array(gtree_array *gt3, int repeat) {
  int k;
  double sumval = 0;
#pragma omp barrier
  for (k=0; k<repeat; k++) {
    sumval += gtree_array_sumval(gt3);
  }
#pragma omp barrier
  return sumval;
}

/* Perform several times the call to the subroutine
   'gtree_array_breadthfirst_sumval'. The number of times 
   is parameterized by 'repeat' parameter.
*/ 
double repeat_breadthfirst_sum_gtree_array(gtree_array *gt3, int repeat, queue_gt3 *queue3) {
  int k;
  double sumval = 0;
#pragma omp barrier
  for (k=0; k<repeat; k++) {
    sumval += gtree_array_breadthfirst_sumval(gt3,queue3);
  }
#pragma omp barrier
  return sumval;
}

/* Perform several times the call to the subroutine
   'gtree_linked_breadthfirst_sumval'. The number of times 
   is parameterized by 'repeat' parameter.
*/ 
double repeat_breadthfirst_sum_gtree_linked(gtree_linked *gt1, int repeat, queue_gt1 *queue1) {
  int k;
  double sumval = 0;
#pragma omp barrier
  for (k=0; k<repeat; k++) {
    sumval += gtree_linked_breadthfirst_sumval(gt1, queue1);
  }
#pragma omp barrier
  return sumval;
}


/* Benchmarks of several algorithms
   and data structures used to implement
   general tree.
*/
int main(int argc, char **argv) {
  /* Number of nodes stored in a single tree */
  const long long int Nmax    = TSIZE;
  /* Four variables that stores the sum
     of values stored into four trees */
  double sumval_linkedlist = 0;
  double sumval_darray = 0; 
  double sumval_breadthfirst = 0;
  double sumval_copy = 0;
#pragma omp parallel reduction(+:sumval_linkedlist,sumval_darray,sumval_breadthfirst,sumval_copy) 
  {
    /* Two dynamic arrays storing two general trees */
    gtree_array    copygt3 = {0};
    gtree_array    gt3 = {0};
    /* One general tree of type gtree_linked */
    gtree_linked      gt1 = {{0},0};
    /* Two queues used to perform breadthfirst algorithm */
    queue_gt3   queue3;
    queue_gt1   queue1;
    /* Number of Bytes stored in a data structure */
    long long int NbBytes;
    /* Number of times we perform the same algorithm
       in order to precisely estimate the execution time 
       of one call */
    long long int NbTimes=REPEAT;
    /* String that stores an output to print */
    char message[50];

    /* Fill two trees 'gt1' and 'gt3' of types 'gtree_linked'
       and 'gtree_array' respectively. The tree 'gt1' is 
       a set of nodes linked together thanks to pointers.
       The tree 'gt3' is an array of nodes linked 
       together with logical indices refering to
       a position in 'gt3->array'.
    */
    printf("You may wait for several seconds ...\n");
    GETTIME(alloc_fill_gtree_array(&gt3, Nmax),"fill_gtree_array");
    printf("You may wait for several seconds (longer than the first time) ...\n");
    GETTIME(alloc_fill_gtree_linked(&gt1, Nmax,&queue1),"fill_gtree_linked");

    /* Perform several times a breadthfirst algorithm to 
       compute the sum of values stored in the nodes of
       tree (data type gtree_linked).
       The total execution time and the bandwidth is
       printed out at the end. */
    NbBytes = Nmax*sizeof(double);
    sprintf(message,"sum_gtree_linked, Mem size=%lld KB",NbBytes/1024);
    GETBWIDTH(sumval_linkedlist=
	      repeat_breadthfirst_sum_gtree_linked(&gt1,NbTimes,&queue1),
	      NbBytes*NbTimes,message);

    /* Perform several times a breadthfirst algorithm to 
       compute the sum of values stored in the nodes of
       tree (data type gtree_array).
       The total execution time and the bandwidth is
       printed out at the end. */
    NbBytes = Nmax*sizeof(double);
    sprintf(message,"sum_breadthfirst_gtree_array, Mem size=%lld KB",NbBytes/1024);
    GETBWIDTH(sumval_breadthfirst=
	      repeat_breadthfirst_sum_gtree_array(&gt3,NbTimes,&queue3),
	      NbBytes*NbTimes,message);

    /* GETTIME(gtree_array_breadthfirst_traversal(&gt3,&queue3),
       "breadthfirst_traversal_trees"); */

    /* Copy the tree (data type gtree_array) to a new tree 'copygt3'.
       The aim of the copy is to reorder nodes such as
       a linear reading of the new array of nodes (copygt3->array)
       corresponds to a breadthfirst order traversal */
    GETTIME(gtree_array_breadthfirst_copy(&gt3,&copygt3,&queue3),
	    "breadthfirst_copy");
 
    /*    GETTIME(gtree_array_traversal(&copygt3),"traversal_trees"); */

    /* Perform AGAIN several times a array traversal algorithm 
       to compute the sum of values stored in the nodes of
       tree (data type gtree_array).
       The total execution time and the bandwidth is
       printed out at the end. */
    NbBytes = Nmax*sizeof(double);
    printf("Sizes %d %d\n",gt3.size,copygt3.size);
    sprintf(message,"sum_ordered_gtree_array, Mem size=%lld KB",NbBytes/1024);
    GETBWIDTH(sumval_copy=repeat_sum_gtree_array(&copygt3,NbTimes),
	      NbBytes*NbTimes,message);

    sprintf(message,"sum_gtree_array, Mem size=%lld KB",NbBytes/1024);
    GETBWIDTH(sumval_darray=repeat_sum_gtree_array(&gt3,NbTimes),
	      NbBytes*NbTimes,message);
  }
  /* Print the values of the sums obtained on the previous trees.
     This is useful to avoid compiler optimization that
     eventually suppress calls to functions that are detected
     as unuseful. */ 
  printf("sums = %.16e %.16e %.16e %.16e\n",sumval_darray, sumval_linkedlist, 
	 sumval_breadthfirst, sumval_copy);
  return 0;
}
