#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "haar.h"

int wav_log2(int input) {
  return((int)round(log2(input)));
}

void inline wav_getmasks(int nblevel, int *masks) {
  int i, j;
  i = (-1)<<1;
  for (j=0; j<=nblevel; j++, i<<=1) { 
    masks[j] = (-1)^i; 
    printf("mask %x\n",masks[j]);
  }
}

int inline wav_level(int input, int nblevel) {
  int j, lev;
  j = 0x1<<nblevel;
  for(j |= input, lev = nblevel-1; ((j&0x1) == 0); j>>=1, lev--);
  return(lev);
}

int inline wav_ilevel(int input, int nblevel) {
  int j = 0x1<<nblevel, lev;
  for(j |= input, lev = 0; 
      ((j&0x1) == 0); j>>=1, lev++);
  return(lev);
}

int inline wav_itf(int input, int nblevel) {
  int j = 0x1<<nblevel, itf;
  for(j |= input, itf=1; 
      ((j&0x1) == 0); j>>=1, itf<<=1);
  return(itf);
}

void inline wav_getmask(int itc, int *itcmask) {
  int i = (-1), j;
  for (j=1; j<itc; j*=2, i<<=1);
  *itcmask = i;
}

/*
  Predict step of the Haar wavelet Scheme.
  
  The term "predict" comes from the articles
  'Building Your Own Wavelets at Home' from Sweldens

  - Transform

  In the wavelet transform, calculate the difference between 
  an odd element and its approximation-interpolation with 
  even coarser points. 
   
  - Inverse Transform

  Reverse the transform predict step by adding the coarse approximation
  (even element) to the difference-detail (odd element).

  Parameters
  - fwav        : the input data on which the forward or inverse
                  transform is calculated.
  - dir(ection) : (1) forward or (-1) inverse transform

*/
void wav_predict(double* fwav, int itc, int N, int dir ) {
  int i, idetail;
  int itf = itc / 2;
  double predictVal;

  for (i = 0; i < N; i += itc) {
    /*  Prediction of a value thanks to an interpolation
	using Fourth-order Lagrange polynomials */
    predictVal = fwav[i];
    /* At index idetail is the detail we
       are looking at */
    idetail = i + itf;
    /* Add or retrieve the predicted value whether
       the forward or inverse transform is applied */
    fwav[idetail] = fwav[idetail] - dir * predictVal;
  } 
} 

void wav_update(double* fwav, int itc, int N, int dir ) {
  int i, idetail;
  int itf = itc / 2;
  double updateVal;

  for (i = 0; i < N; i += itc) {
    /* At index idetail is the detail we
       are looking at */
    idetail = i + itf;

    updateVal = .5 * fwav[idetail];

    /* Add or retrieve the update value to the even
       points whether the forward or inverse transform 
       is applied */
    fwav[i] = fwav[i] + dir * updateVal;
  } 
} 

/*
  Wavelet Haar scheme transform.
  The wavelet Haar scheme is an in-place wavelet algorithm.
  A series of N elements is passed to the transform
  function.  The result of the wavelet lifting scheme is calculated
  in place (without any array temporaries) in 'fwav'
*/
void wav_ftransform( double *fwav, int N ) {
  int j;
  for (j = 2; j <= N; j = j * 2) {
    wav_predict( fwav, j, N, 1 );
    wav_update( fwav, j, N, 1 );
  } 
} 

/*
  Predict step of the wavelet Haar Scheme.
*/
void wav_adapt_predict(FTYPE *sdata, int itc, 
		       int N, int dir) {
  _KEYTYPE i, idetail;
  int      itf = itc / 2;
  double   predictVal, detailVal, *pdetail;
#ifdef _HASH
  int      p, pstart, pend, ilevel;
  ilevel = wav_log2(itc);
  if (ilevel == 0) pstart = 0.;
  else             pstart = sdata->ends[ilevel-1];
  pend   = sdata->cursize;
  for (p=pstart; p<pend; p++) {
    i          = sdata->keys[p];
    predictVal = sdata->vals[p];
#else
  for (i = 0; i < N; i += itc) {
    _GETF(sdata,i,predictVal);
#endif
    /* detail at index idetail*/
    idetail = i + itf;
    _GETF(sdata,idetail,detailVal);
    /* Verify if detail or coarse is NOTAVALUE */
    if ((predictVal == NOTAVALUE) && 
	(detailVal != NOTAVALUE)) {
      printf("!(perfect tree) in adapt_pred"\
	     "i %d itc %d\n",(int)i,(int)itc);
      exit(2);
    }
    if (detailVal != NOTAVALUE) {
      _GETPOINTER(sdata,idetail,pdetail);
      *pdetail -= dir * predictVal;
    }}} 

/*
  Update step of the adaptive wavelet Haar Scheme.
*/
void wav_adapt_update( FTYPE *swav, int itc, int N, int dir ) {
  _KEYTYPE i, idetail;
  int itf = itc / 2;
  double predictVal, detailVal;
  double *pval;

#ifdef _HASH
  int p, pstart, pend, ilevel;
  ilevel = wav_log2(itc);
  if (ilevel == 0) pstart = 0;
  else             pstart = swav->ends[ilevel-1];
  pend   = swav->cursize;
  for (p=pstart; p<pend; p++) {
    /*  At index i we get a prediction of a value  */
    i          = swav->keys[p];
    predictVal = swav->vals[p];
    pval       = &(swav->vals[p]);
#else
  for (i = 0; i < N; i += itc) {
    /*  At index i we get a prediction of a value  */
    _GETF(swav,i,predictVal);
    _GETPOINTER(swav,i,pval);
#endif
    
    /* At index idetail is the detail we
       are looking at */
    idetail = i + itf;
    _GETF(swav,idetail,detailVal);

    /* Take into account that NOTAVALUE could be
       set in locations 'i' ou 'idetail' */
    if ((predictVal == NOTAVALUE) && (detailVal != NOTAVALUE)) {
      printf("Not a perfect tree in wav_adapt_update i %d idetail %d itc %d pVal %lf\n",
	     (int)i,(int) idetail, (int)itf, predictVal);
      exit(2);
    }
    if (detailVal != NOTAVALUE) {
      /* Add or retrieve the predicted value whether
	 the forward or inverse transform is applied */
      *pval = *pval + dir * .5 * detailVal;
    }}} 


/*
  Wavelet lifting scheme transform.
  
  The wavelet lifting scheme is an in-place wavelet algorithm.
  A series of N elements is passed to the <tt>transform</tt>
  function.  The result of the wavelet lifting scheme is calculated
  in place (without any array temporaries) in 'ts'
*/
void wav_adapt_ftransform(FTYPE *sdata, int N) {
  int j;
  for (j = 2; j <= N; j = j * 2) {
    wav_adapt_predict(sdata, j, N, 1);
    wav_adapt_update(sdata, j, N, 1);
  } 
} 

/*
  Wavelet inverse transform.
*/
void wav_itransform( double *fwav, int N ) {
  int j;
  /* Perform the reconstruction */
  for (j = N; j >= 2; j = j / 2) {
    wav_update( fwav, j, N, -1);
    wav_predict( fwav, j, N, -1);
  }
}

/*
  Wavelet inverse transform.
*/
void wav_adapt_itransform( FTYPE *swav, int N ) {
  int j;
  /* Perform the reconstruction */
  for (j = N; j >= 2; j = j / 2) {
    wav_adapt_update( swav, j, N, -1);
    wav_adapt_predict( swav, j, N, -1);
  }
}


void wav_reorder_grid( FTYPE *sold, FTYPE *snew, int N ) {
#ifdef _HASH
  int p, i;
  _KEYTYPE key;
  _VALTYPE val;
  int nblevel, ilevel;
  int *hole;
  _RESETF(snew,N);
  nblevel=wav_log2(N);
  hole = (int*)malloc((2+nblevel)*sizeof(int));
  for (ilevel=0; ilevel<2+nblevel; ilevel++) {
    hole[ilevel] = 0;
  }
  for (p=0; p<sold->cursize; p++) {
    ilevel = wav_ilevel(sold->keys[p],nblevel);
    hole[ilevel]++;
  }
  snew->ends[0] = hole[0];
  for (ilevel=1; ilevel<1+nblevel; ilevel++) {
    snew->ends[ilevel] = hole[ilevel] + snew->ends[ilevel-1];
  }
  hole[0]=0;
  for (ilevel=1; ilevel<1+nblevel; ilevel++) {
    hole[ilevel] = snew->ends[ilevel-1];
  }
  

  for (p=0; p<sold->cursize; p++) {
    key = sold->keys[p];
    val = sold->vals[p];
    ilevel = wav_ilevel(key,nblevel);
    i = hole[ilevel];
    _SETF_MODIFY(snew,key,val,i);
    hole[ilevel]++;
  }
//  for (ilevel=0; ilevel<1+nblevel; ilevel++) 
//    printf("reorder %d hole %d ilev %d\n",
//	   ilevel,hole[ilevel],snew->ends[ilevel]);
  free(hole);
#endif
    
}

/*
  Thresholding of coefficients
*/
void wav_thresholding( double *fwav, int N, double normcoef , double threshold, double *fold ) {
  int itc, itf, itv, nnz_tot, i;
  register int cond;
  double *threshold_level;
  int *nnz_level;
  int nblevel, level;
  nblevel = wav_log2(N);
  threshold_level = (double*) malloc (nblevel*sizeof(double));
  nnz_level = (int*) malloc (nblevel*sizeof(int));
  for (level = 0; level < nblevel; level++) {
    itc = pow(2,nblevel-level);
    threshold_level[level] = threshold * pow(normcoef, level);
    nnz_level[level] = 0;
  }
  nnz_tot = 1; /* counting coefficient of index 0 */
  level = nblevel-1;
  /* Look at finest details */
  for (i = 1; i < N; i +=2) {
    if (fabs(fwav[i]) < threshold_level[level]) {
      fwav[i] = NOTAVALUE;
      fold[i] = NOTAVALUE;
    } else {
      nnz_level[level]++;      
    }
  }
  /* Look at details of coarser levels */
  for (level = nblevel-2, itf = 2, itc = 4; itc <= N; 
       itc *= 2, itf *= 2, level--) {
    for (i = itf; i < N; i +=itc) {
      cond = 1; 
      if (fabs(fwav[i]) < threshold_level[level]) {
	for (itv = 1; itv <= itf/2; itv *=2) 
	  cond &= (fwav[i+itv] == NOTAVALUE);
      } else {
	cond = 0;
      }
      if (cond) {
	fwav[i] = NOTAVALUE;
	fold[i] = NOTAVALUE;
      } else {
	nnz_level[level]++;      
      }
    }
  }
  for (level = 0; level < nblevel; level++) {
//    WAVDEBUG(printf("\nlevel %d N %d threshold %e nnz %d\n",
//		    level,itc,threshold_level[level],nnz_level[level]));
    nnz_tot += nnz_level[level];
  }
  printf("Number of non-zero coefficients :"\
	 "%d over %d (%.3f percents)\n",nnz_tot,N,(100.*nnz_tot)/N);
  free(nnz_level);
  free(threshold_level);
}

/*
  Thresholding of coefficients
*/
void wav_adapt_thresholding( FTYPE *swav, int N, double normcoef , double threshold) {
  int itc, itf, itv, nnz_tot;
  _KEYTYPE i, idet;
  register int cond;
  double *threshold_level;
  double *pval;
  double val, valk;
  int *nnz_level;
  int nblevel, level, ilevel;
#ifdef _HASH
  int pstart, pend, p;
#endif
  nblevel = wav_log2(N);
  threshold_level = (double*) malloc (nblevel*sizeof(double));
  nnz_level = (int*) malloc (nblevel*sizeof(int));
  for (level = 0; level < nblevel; level++) {
    itc = (int)round(pow(2,nblevel-level));
    threshold_level[level] = threshold * pow(normcoef, level);
    nnz_level[level] = 0;
  }
  nnz_tot = 1; /* counting coefficient of index 0 */
  level = nblevel-1;
  /* Look at finest details */
#ifdef _HASH
  pstart = 0;
  pend   = swav->ends[0];
  for (p=pstart; p<pend; p++) {
    i   = swav->keys[p];
    val = swav->vals[p];
#else
  for (i = 1; i < N; i +=2) {
    _GETF(swav,i,val);
#endif
    if ((val != NOTAVALUE) && (fabs(val) < threshold_level[level])) {
      val = NOTAVALUE;
      _GETPOINTER(swav,i,pval); *pval = NOTAVALUE;
      //      _GETPOINTER(sold,i,pval); *pval = NOTAVALUE;
    } 
    if (val != NOTAVALUE) {
      nnz_level[level]++;      
    }
  }
  /* Look at details of coarser levels */
  for (ilevel = 1, level = nblevel-2, itf = 2, itc = 4; itc <= N; 
       itc *= 2, itf *= 2, level--, ilevel++) {
#ifdef _HASH
    pstart = pend;
    pend   = swav->ends[ilevel];
    for (p=pstart; p<pend; p++) {
      i   = swav->keys[p];
      val = swav->vals[p];
#else
    for (i = itf; i < N; i +=itc) {
      _GETF(swav,i,val);
#endif
      cond = 1; 
      /* Scan finer levels to build a correct wavelet tree */
      if ((val != NOTAVALUE) && (fabs(val) < threshold_level[level])) {
	for (itv = 1; itv <= itf/2; itv *=2) {
	  idet = i + itv;
	  _GETF(swav,idet,valk);
	  cond &= (valk == NOTAVALUE);
	}
      } else {
	cond = 0;
      }
      if (cond) {
	val = NOTAVALUE;
	_GETPOINTER(swav,i,pval); *pval = NOTAVALUE;
	//	_GETPOINTER(sold,i,pval); *pval = NOTAVALUE;
      } 
      if (val != NOTAVALUE) {
	nnz_level[level]++;      
      }
    }
  }
  for (level = 0; level < nblevel; level++) {
//    WAVDEBUG(printf("\nlevel %d N %d threshold %e nnz %d\n",
//		    level,itc,threshold_level[level],nnz_level[level]));
    nnz_tot += nnz_level[level];
  }
  printf("Number of non-zero coefficients :"\
	 "%d over %d (%.3f percents)\n",nnz_tot,N,(100.*nnz_tot)/N);
  free(nnz_level);
  free(threshold_level);
}

/*
  Count Number of Non-Zero coefficients
*/
int wav_getnnz( FTYPE *swav, int N ) {
  int nnz_tot;
  _VALTYPE val;
#ifdef _HASH
  int p;
  nnz_tot = 0;
  for (p=0; p<swav->cursize; p++) {
    val = swav->vals[p];
#else
  _KEYTYPE i;
  nnz_tot = 0;
  for (i = 0; i < N; i++) {
    _GETF(swav,i,val);
#endif
    if (val != NOTAVALUE) nnz_tot++;
  }
  return(nnz_tot);
}

/*
  Zeroing NOTAVALUE values
*/
void wav_zeroing( double *fold, int N ) {
  register int i;
  for (i = 0; i < N; i ++) 
    if (fold[i] == NOTAVALUE) 
      fold[i] = 0.;
}


/*
  Copy an array from fold to fnew
*/
void wav_copyfunc_aa( double *fold, double *fnew, int N ) {
  register int i;
  for (i = 0; i < N; i ++) fnew[i] = fold[i];
}

/*
  Copy an array from fold to snew
*/
void wav_copyfunc_as( double *fold, FTYPE *snew, FTYPE *stmp, int N) {
  register int i;
  register double val;
  FTYPE           *starget;
#ifdef _HASH
  starget = stmp;
#else
  starget = snew;
#endif
  _RESETF(starget,N);
  for (i = 0; i < N; i ++) {
    val = fold[i];
    if (val != NOTAVALUE){
      //      printf("copy %d  = %f\n",i,val);
      _SETF_NEW(starget,i,val);
    }
  }
#ifdef _HASH
  wav_reorder_grid(stmp,snew,N);
//  { 
//    int ilevel, nblevel = wav_log2(N);
//    for (ilevel=0; ilevel<1+nblevel; ilevel++) 
//      printf("copy %d  = %d\n",ilevel,
//	     snew->ends[ilevel]);
//  }
#endif
}

/*
  Copy an array from fold to snew
*/
void wav_copyfunc_ss( FTYPE *sold, FTYPE *snew, int N) {
  _VALTYPE val;
#ifdef _HASH
  int ilevel, nblevel, p;
  _KEYTYPE key;
  _RESETF(snew,N);
  for (p=0; p<sold->cursize; p++) {
    key = sold->keys[p];
    val = sold->vals[p];
    _SETF_MODIFY(snew,key,val,p);
  }
  nblevel = wav_log2(N);
  for (ilevel=0; ilevel<nblevel+1; ilevel++) {
    snew->ends[ilevel] = sold->ends[ilevel];
  }
#else
  int i;
  _RESETF(snew,N);
  for (i = 0; i < N; i ++) {
    _GETF(sold,i,val);
    if (val != NOTAVALUE){
      _SETF_NEW(snew,i,val);
    }
  }
#endif
}

/*
  Copy an array from fold to snew
*/
void wav_copyfunc_sa( FTYPE *sold, double *fnew, int N) {
  int i;
  register double val;
  for (i = 0; i < N; i ++) {
    _GETF(sold,i,val);
    fnew[i] = val;
  }
}

/*
  Copy an array from fold to snew
*/
void wav_review( FTYPE *sold, int N ) {
  int i;
  register double val;
  for (i = 0; i < N; i ++) {
    _GETF(sold,i,val);
    if (val != NOTAVALUE){
      printf("val %d %f\n",i,val);
    }
  }
}


/*
  Advect input signal grid (back and forth)
*/
void wav_advect_grid( FTYPE *swav, int N, double displacement, FTYPE *snew, FTYPE *stmp ) {
  int k, j, c, ilevel, nblevel;
  int itc, itf, atf, floordisp, intdisp, maskatf;
  int i, beginpatch, patchsize, icoa;
  double val, valdet;
  double *pval;
#ifdef _HASH
  int p, pstart, pend, myitf;
  pend = 0;
#endif
  nblevel = wav_log2(N);
  patchsize = 3;

  floordisp = (int)round(floor(displacement));
  _RESETF(snew,N);
  
  /* scan all levels */
  for (itf = 1, itc = 2, ilevel = 0; itc <= N; 
       itf *=2, itc*=2, ilevel++) {
    atf = itf/2; if (atf < 1) atf = 1;
    /* higher bits mask corrresponding to atf */
    wav_getmask(atf, &maskatf);
    /* traversal at level 'itc' */
#ifdef _HASH
    pstart = pend;
    pend   = swav->ends[ilevel];
    for (p=pstart; p<pend; p++) {
      i   = swav->keys[p];
      val = swav->vals[p];
#else
    for (i = itf; i < N; i+= itc) { 
      _GETF(swav,i,val);
#endif
      if (val != NOTAVALUE) {
	/* advect grid point */
	intdisp    = (i + floordisp + N)%N;
	/* get a 'divide atf' grid point */
	intdisp   &= maskatf;
	/* SAME AS: intdisp /= atf; intdisp *= atf; */
	/* refinement procedure, insert 
	   patchsize points for each initial point */
	beginpatch = N+intdisp-atf;
	for (c = 0; c < patchsize; c++) {
	  j = (beginpatch + c*atf)%N;
	  _GETF(snew,j,val);
	  if (val == NOTAVALUE) {
	    _SETF_NEW(snew,j,0.);
	  }}}}}
  /* Scan levels to build a correct wavelet tree */
  for (itf = 1, itc = 2; itc <= N; itf *=2, itc*=2) {
#ifdef _HASH
    for (p=0; p<snew->cursize; p++) {
      i      = snew->keys[p];
      valdet = snew->vals[p];
      myitf=wav_itf(i,nblevel);
      if ((myitf == itf) && (valdet != NOTAVALUE)) {
#else
    for (i = itf; i < N; i+= itc) { 
      _GETF(snew,i,valdet);
      if (valdet != NOTAVALUE) {
#endif
	icoa = i-itf;
	_GETF(snew,icoa,val);
	if (val == NOTAVALUE) {
 	_SETF_NEW(snew,icoa,0.);
	}
      }
    }
  }
#ifdef _HASH
  wav_reorder_grid(snew,stmp,N);
  wav_copyfunc_ss(stmp,snew,N);
#endif

#ifdef _HASH
  for (p=0; p<snew->cursize; p++) {
    i = snew->keys[p];
    val = snew->vals[p];
#else
  for (i = 0; i < N; i +=1) {
    _GETF(snew,i,val);
#endif
    if (val != NOTAVALUE) {
//OLD      maxitf = wav_itf(i, nblevel);
//OLD      for (itf=1, k=intdisp; (itf < maxitf) && (valk == NOTAVALUE); itf*=2) {	
//OLD 	k &= (-1)^itf;
//OLD	_GETF(swav,k,valk);
//OLD      }

      k = (i - floordisp + N)%N;
      _GETPOINTER(swav,k,pval);
      
      if (pval == NULL) {
	/* Add a grid point at the same level as input point*/
	_SETF_NEW(swav,k,0.);
      }	else if (*pval ==NOTAVALUE) {
	*pval=0.;
      } 
     }
   }

    /* Scan all levels to build a correct wavelet tree */
  for (itf = 1, itc = 2; itc <= N; itf *=2, itc*=2) {
#ifdef _HASH
    for (p=0; p<swav->cursize; p++) {
      i      = swav->keys[p];
      valdet = swav->vals[p];
      myitf=wav_itf(i,nblevel);
      if ((myitf == itf) && (valdet != NOTAVALUE)) {
#else
    for (i = itf; i < N; i+= itc) { 
      _GETF(swav,i,valdet);
      if (valdet != NOTAVALUE) {
#endif
	icoa = i-itf;
	_GETPOINTER(swav,icoa,pval);
        if (pval == NULL) {
	   /* Add a grid point at the same level as input point*/
	   _SETF_NEW(swav,icoa,0.);
        } else  if (*pval ==NOTAVALUE) {
	   *pval=0.;
        } 
      }
    }
  }
#ifdef _HASH
  wav_reorder_grid(swav,stmp,N);
  wav_copyfunc_ss(stmp,swav,N);
#endif

}

 

 
/*
  Advect input signal, the grid is already prepared in the snew array
*/
void wav_advect( FTYPE *swav, int N, double displacement, FTYPE *snew) {
  int i, k;
  double valk, val, *pval;
  int floordisp;
#ifdef _HASH
  int p;
#endif
  floordisp = (int)round(floor(displacement));
#ifdef _HASH
  for (p=0; p<snew->cursize; p++) {
    i = snew->keys[p];
    val = snew->vals[p];
#else
  for (i = 0; i < N; i +=1) {
    _GETF(snew,i,val);
#endif
    if (val != NOTAVALUE) {
      k = (i - floordisp + N)%N;
      _GETF(swav,k,valk);
      
      if (valk != NOTAVALUE) {
	/* Copy the value of the function 'valk' */
	_GETPOINTER(snew,i,pval);
	*pval = valk;
      } else {
 	printf("Wavelet support not found i %d k %d\n",i,k);
 	exit(2);
      }
     }
   }
 }



