#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "clock.h"
#include "haar.h"
#ifndef DRATE
#define DRATE 1000000000
#endif
#ifndef TSIZE
#define TSIZE 1024*1024
#endif

/* Write into a file the contents of the array "data" */
/* The name of the file is "fname", the "size" specifies the array size */
void print1d(double *data, int size, double dx, char *fname) {
  FILE*  fd;
  register int i;
  register double val;
  if ((fd=fopen(fname,"w")) == NULL) {
    printf("file %s could not be opened\n",fname);
    exit(-1);
  }
  for (i = 0; i < size; i++) {
    val = data[i];
    fprintf(fd,"%f %.3e\n",dx*i,val);
  }
  fclose(fd);
}

/* Write into a file the contents of the array "data" except NOTAVUE values. */
/* The name of the file is "fname", the "size" specifies the array size */
void wprint1d(double *data, int size, double dx, char *fname) {
  FILE*  fd;
  int i;
  register double val;
  if ((fd=fopen(fname,"w")) == NULL) {
    printf("file %s could not be opened\n",fname);
    exit(-1);
  }
  for (i = 0; i < size; i++) {
    val = data[i];
    if (val != NOTAVALUE)
      fprintf(fd,"%f %.3e\n",dx*i,val);
  }
  fclose(fd);
}

/* Write into a file the contents of the array "FTYPE *data" except NOTAVUE values. */
/* The name of the file is "fname", the "size" specifies the array size */
void sparse_print1d_opt(FTYPE *data, int size, double dx, char *fname) {
  FILE*             fd;
  register _KEYTYPE i;
  register _VALTYPE val;
  if ((fd=fopen(fname,"w")) == NULL) {
    printf("file %s could not be opened\n",fname);
    exit(-1);
  } else {
#ifdef _HASH
    register int      p;
    for (p=0; p<data->cursize; p++) {
      val = data->vals[p];
      i   = data->keys[p];
#else
    for (i = 0; i < size; i++) {
      _GETF(data,i,val);
#endif  
      if (val != NOTAVALUE)
	fprintf(fd,"%f %.3e\n",dx*i,val);
    }
  }
  fclose(fd);
}


void sparse_print1d(FTYPE *sdata, int size, 
		    double dx, char *fname) {
  FILE*    fd;
  _KEYTYPE i;
  _VALTYPE val;
  if ((fd=fopen(fname,"w")) == NULL) {
    printf("file %s could not be opened\n",
	   fname);
    exit(-1);
  } else {
    for (i = 0; i < size; i++) {
      _GETF(sdata,i,val);
      if (val!= NOTAVALUE)
	fprintf(fd,"%f %.3e\n",dx*i,val);
    }
  }
  fclose(fd);
}

/* Write into a file the contents of the array "FTYPE* data". */
/* The name of the file is "fname", the "size" specifies the array size */
void waprint1d(FTYPE *data, int size, double dx, char *fname) {
  FILE*  fd;
  int i;
  register double val;
  if ((fd=fopen(fname,"w")) == NULL) {
    printf("file %s could not be opened\n",fname);
    exit(-1);
  }
  for (i = 0; i < size; i++) {
    _GETF(data,i,val);
      fprintf(fd,"%f %.3e\n",dx*i,val);
  }
  fclose(fd);
}

/* Write into a file the contents of the array "data" */
/* The name of the file is "fname", the "size" specifies the array size */
void average1d(double *data, int size, char *mess) {
  register int i;
  register double sum;
  sum = 0;
  for (i = 0; i < size; i++) {
    sum += data[i];
  }
  printf("%s %e\n",mess,sum/(float)size);
}

/* Initialize the initial array with a sine function 
   multiplied by a decreasing exp(-x) coefficient */
  /* Retrieve the value of 2*PI */
void init_func1(double *data, int size) {
  double  dpi, coeff, dx, val, maxval, inv_maxval, ftmp;
  int     modex, i;
  /* Two times PI */
  dpi   = 8*atan(1.);
  /* Excited mode */
  modex = 6;
  /* Mesh size */
  dx    = dpi/size;
  coeff = dx*modex;
  /* Initialize random generator */
  srand48(4);
  /* Fill "data" array with a signal + noise */
  /* The signal is a sine times an exponentional decrease */
  maxval = 0;
  for (i = 0; i < size; i++) {
    ftmp = exp(-dx*2*abs(i-size/2));
    ftmp *= sin(coeff*i);
    val =  ftmp + .001*drand48();
    data[i] = val;
    val = fabs(val);
    maxval = (val > maxval?val:maxval);
  }
  inv_maxval = 1./maxval;
  for (i = 0; i < size; i++) data[i] = data[i] * inv_maxval;
}

/* Program that runs a haar wavelet scheme
   on a 1D data set : a noisy sine signal */
int main(int argc, char **argv) {
  double  *f_old, *f_wav, *f_tmp;
  FTYPE   *s_old, *s_new, *s_wav;
  int     Nmax = TSIZE;
  int     nnztot;
  //  double  Lp = 100000;
  double  normcoef = 1.;//pow(2,-1/Lp);
  double  thresholdMax = .003;
  int     i;
  double  displacement;
  double  dx;
  int     nbdisplacements;
  char    message[50];

  f_old = (double*)malloc(Nmax*sizeof(double));
  f_wav = (double*)malloc(Nmax*sizeof(double));
  f_tmp = (double*)malloc(Nmax*sizeof(double));
  _INITF (&s_new,Nmax);
  _INITF (&s_wav,Nmax);
  _INITF (&s_old,Nmax);
  printf("Nmax = %d\n",Nmax);

  dx    = atan(8.)/Nmax;
  /* Initialize initial array with a modified sine signal */
  GETTIME(init_func1(f_old,Nmax),"init_func1:");
  nbdisplacements = 32;
  displacement = Nmax/nbdisplacements;
  /* print input signal into a file 'signal' */
  WAVDEBUG(print1d(f_old,Nmax,dx,"input.dat"));
  average1d(f_old,Nmax,"average of input:");

//  /*==========  UNITARY TEST1  ==========*/
//  /* Copy signal from f_old to f_wav */
//  GETTIME(wav_copyfunc_aa(f_old,f_wav,Nmax),"fcopy_A:");
//  /* Wavelet forward transform */
//  GETTIME(wav_ftransform(f_wav,Nmax),"ftransf:");
//  /* Thresholding of details */
//  GETTIME(wav_thresholding(f_wav,Nmax,normcoef,thresholdMax,f_old),"threshold:");
//  GETTIME(wav_zeroing(f_wav,Nmax),"zeroing:");
//  /* Wavelet iverse transform */
//  GETTIME(wav_itransform(f_wav,Nmax),"ftransf:");
//  average1d(f_wav,Nmax,"average of input:");
//  WAVDEBUG(print1d(f_wav,Nmax,dx,"output.dat"));
//  exit(0);
//  /*=====================================*/
//  /*==========  UNITARY TEST2  ==========*/
//  /* Copy signal from f_old to f_wav */
//  GETTIME(wav_copyfunc_aa(f_old,f_wav,Nmax),"fcopy_A:");
//  /* Wavelet forward transform */
//  GETTIME(wav_ftransform(f_wav,Nmax),"ftransf:");
//  /* Thresholding of details */
//  GETTIME(wav_thresholding(f_wav,Nmax,normcoef,thresholdMax,f_old),"threshold:");
//  /* Copy signal from f_wav to s_new */
//  GETTIME(wav_copyfunc_as(f_wav,s_new,s_wav,Nmax),"fcopy:");
//  GETTIME(wav_adapt_itransform(s_new,Nmax),"adapt_itransf:");
//  GETTIME(wav_adapt_ftransform(s_new,Nmax),"adapt_ftransf:");
//  GETTIME(wav_copyfunc_sa(s_new,f_tmp,Nmax),"fcopy:");
//  GETTIME(wav_zeroing(f_tmp,Nmax),"zeroing:");
//  GETTIME(wav_itransform(f_tmp,Nmax),"itransf:");
//  average1d(f_tmp,Nmax,"average of input:");
//  exit(0);
//  /*=====================================*/


  /* Copy signal from f_old to f_wav */
  GETTIME(wav_copyfunc_aa(f_old,f_wav,Nmax),"fcopy:");
  /* Wavelet forward transform on f_wav */
  GETTIME(wav_ftransform(f_wav,Nmax),"ftransf:");
  /* Thresholding of details on f_wav  (put NOTAVALUE also in f_old as needed) */
  GETTIME(wav_thresholding(f_wav,Nmax,normcoef,thresholdMax,f_old),
	  "threshold:");
  wprint1d(f_old,Nmax,dx,"sparsef_beg.dat");

  /* Copy signal from f_wav to f_tmp */
  GETTIME(wav_copyfunc_aa(f_wav,f_tmp,Nmax),"fcopy:");
  /* Put zero instead of NOTAVALUE in 'f_tmp' */
  GETTIME(wav_zeroing(f_tmp,Nmax),"zeroing:");
  /* Wavelet inverse transform on f_tmp */
  GETTIME(wav_itransform(f_tmp,Nmax),"itransf:");
  average1d(f_tmp,Nmax,"average of input:");
  sprintf(message,"out_beg.dat");
  WAVDEBUG(print1d(f_tmp,Nmax,dx,message));


  /* Copy signal from f_wav to s_wav to initiate the next loop */
  GETTIME(wav_copyfunc_as(f_wav,s_wav,s_old,Nmax),"fcopy_A:");

  for (i = 1; i <= nbdisplacements; i++) {
    /* Grid Advection : modify both s_wav (old signal) and s_new (new signal) */
    GETTIME(wav_advect_grid(s_wav,Nmax,-displacement,s_new,s_old),"advect_grid:");
    /* Adaptive wavelet inverse transform */
    GETTIME(wav_adapt_itransform(s_wav,Nmax),"adapt_itransf_A:");
    /* Advection+interpolation of coefficients => 
       s_new contains new advected values */
    GETTIME(wav_advect(s_wav,Nmax,-displacement,s_new),"advect");
 
    /* Count the number of Non-Zero coefficients */
    GETTIME(nnztot=wav_getnnz(s_new,Nmax),"count nnz:");
    printf("NNZ %d\n",nnztot);
    /* Wavelet forward transform in s_new */
    GETTIME(wav_adapt_ftransform(s_new,Nmax),"adapt_ftransf_A:");

    /* copy s_new to s_wav */
    GETTIME(wav_copyfunc_ss(s_new,s_wav,Nmax),"fcopy_B:");
    /* Thresholding of details, s_new contains wavelet coeff. */
    GETTIME(wav_adapt_thresholding(s_wav,Nmax,normcoef,thresholdMax),"threshold:");
    GETTIME(nnztot=wav_getnnz(s_wav,Nmax),"count nnz:");
    printf("NNZ AT THE END OF ITER %d: %d\n",i,nnztot);

    if (i%DRATE == 0) {
      /* Copy signal from s_new to s_wav */
      WAVDEBUG(GETTIME(wav_copyfunc_sa(s_wav,f_tmp,Nmax),"fcopy:"));
      WAVDEBUG(GETTIME(wav_zeroing(f_tmp,Nmax),"zeroing:"));
      WAVDEBUG(GETTIME(wav_itransform(f_tmp,Nmax),"itransf:"));
      /* print restored signal into a file 'output' */
      sprintf(message,"out_%d.dat",i/DRATE);
      printf("out %s\n",message);
      WAVDEBUG(print1d(f_tmp,Nmax,dx,message));
      WAVDEBUG(average1d(f_tmp,Nmax,"average of output:"));
    }
  }
  i=nbdisplacements;
  GETTIME(wav_copyfunc_sa(s_wav,f_tmp,Nmax),"fcopy:");

  GETTIME(wav_adapt_itransform(s_wav,Nmax),"adapt_itransf_end:");
  sparse_print1d(s_wav,Nmax,dx,"sparsef_end.dat");

  /* Put zero instead of NOTAVALUE in 'f_tmp' */
  GETTIME(wav_zeroing(f_tmp,Nmax),"zeroing:");
  /* Wavelet inverse transform on f_tmp */
  GETTIME(wav_itransform(f_tmp,Nmax),"itransf:");
  sprintf(message,"out_end.dat");
  WAVDEBUG(print1d(f_tmp,Nmax,dx,message));
  average1d(f_tmp,Nmax,"average at the end:");
  
  free(f_old);
  free(f_wav);
  free(f_tmp);
  _DESTROY(&s_old);
  _DESTROY(&s_wav);
  _DESTROY(&s_new);
  return 0;
}
