#include "haar.h"
#include <malloc.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
/*
  Predict step of the wavelet Lifting Scheme.
  
  The term "predict" comes from the articles
  'Building Your Own Wavelets at Home' from Sweldens

  - Transform

  In the wavelet transform, calculate the difference between 
  an odd element and its approximation-interpolation with 
  even coarser points. 
   
  - Inverse Transform

  Reverse the transform predict step by adding the coarse approximation
  (even element) to the difference-detail (odd element).

  - Explanation 

  Predict an odd point using Haar approximation.
  The odd point is the average of neighbouring odd and even points.
  
  The difference between the predicted (averaged) value
  and the actual odd value replaces the odd value in the
  forward transform.
  
  As the recursive steps proceed, N will eventually be 4 and
  then 2.  
  
  Parameters
  - vec       : the input data on which the forward or inverse
                transform is calculated.
  - N         : the area of vec over which the transform 
                is calculated
  - direction : (1) forward or (-1) inverse transform

*/
void haar_predict(double* vec, int itc, 
		  int N, int dir) {
  int i, itf = itc / 2;
  for (i = 0; i < N; i += itc) 
    vec[i+itf] -= dir * vec[i];
} 








/*
  The update step of the wavelet Lifting Scheme
  
  In the Lifting Scheme transform the update step follows
  the predict step.  After the predict step, the differences
  have been calculated in-place writing over the even (b)
  values.  
*/
void haar_update( double* vec, int itc, int N, int dir ) {
  int itf = itc / 2;
  int i, idetail;
  double updateVal;
  
  for (i = 0; i < N; i += itc) {
    /* At index idetail is the detail we
       are looking at */
    idetail = i + itf;

    updateVal = .5 * vec[idetail];

    /* Add or retrieve the update value to the even
       points whether the forward or inverse transform 
       is applied */
    vec[i] = vec[i] + dir * updateVal;
  } 
} 


/*
  Wavelet lifting scheme transform.
  
  The wavelet lifting scheme is an in-place wavelet algorithm.
  A series of N elements is passed to the <tt>transform</tt>
  function.  The result of the wavelet lifting scheme is calculated
  in place (without any array temporaries) in 'ts'
*/
void haar_ftransform(double *ts, int N) {
  int itc;
  for (itc = 2; itc <= N; itc *= 2) {
    haar_predict( ts, itc, N, 1 ); 
    haar_update( ts, itc, N, 1 );	
  } 
} 

/*
  Wavelet lifting Scheme inverse transform.
*/
void haar_itransform(double *ts, int N) {
  int itc;
  for (itc = N; itc >= 2; itc /= 2) {
    haar_update( ts, itc, N, -1);  
    haar_predict( ts, itc, N, -1); 
  }
}
int inline haar_level(int input, int nblevel) {
  int j, lev;
  for(j = input, lev = nblevel-1; ((j&0x1) == 0); j>>=1, lev--); return(lev);
}

inline double threshold_func(double threshold, double norm, int level) {
  return(threshold * pow(norm, ((double)level)));
}  

/* Thresholding of coefficients */
void haar_thresholding(double *vec, int N, 
		       double normcoef, 
		       double threshold) {
  int nnz_tot; /* number of non-zero coeff. */
  register int i;
  double *threshold_level;
  /* number of non-zero coeff. for each level */
  int    *nnz_level;
  /*  number of level, actual level */
  int    nblevel, level;
  nblevel         = (int)round(log2(N));
  threshold_level = 
    (double*) malloc (nblevel*sizeof(double));
  nnz_level       = 
    (int*) malloc (nblevel*sizeof(int));
  for (level = 0; level < nblevel; level++) {
    threshold_level[level] = 
      threshold_func(threshold,norm,level);
    nnz_level[level] = 0;
  }
  for (i = 1; i < N; i +=1) {
    level = haar_level(i, nblevel);
    if (fabs(vec[i]) < threshold_level[level]) {
      vec[i] = 0.;
    } else {
      nnz_level[level]++;      
    }
  }
  nnz_tot = 0;
  for (level = 0; level < nblevel; level++) {
    printf("level %4d threshold %20e nnz %10d\n",
	   level,threshold_level[level],
	   nnz_level[level]);
    nnz_tot += nnz_level[level];
  }
  printf("Number of non-zero coefficients :"\
	 "%13d over %13d (%.7f percents)\n",
	 nnz_tot,N,(100.*nnz_tot)/N);
  free(nnz_level);
  free(threshold_level);
}


void haar_fdwt(double* vec, int N) {
  register int itf, itc;
  register int idetail, icoarse;
  /* loop from finest level to coarsest level */
  for (itc = 2, itf = 1; 
       itc <= N; itc *= 2, itf *= 2) {

    /* loop on all coefficients at this level */
    for (icoarse = 0; 
	 icoarse < N; icoarse += itc) {
      /* At index 'idetail': the difference
	      at 'icoarse': the average */
      idetail       = icoarse + itf;
      /* PREDICT */
      vec[idetail]  = 
	vec[idetail] - vec[icoarse];
      /* UPDATE */
      vec[icoarse] = 
	vec[icoarse] + .5 * vec[idetail];
    }
  }
} 

void haar_idwt(double* vec, int N) {
  register int itf, itc;
  register int idetail, icoarse;
  /* loop from coarsest level to finest level */
  for (itc = N, itf = N/2; 
       itc >= 2; itc /= 2, itf /= 2) {
    /* loop on all coefficients at this level */
    for (icoarse = 0; 
	 icoarse < N; icoarse += itc) {

      /* At index 'idetail': the difference
	      at 'icoarse': the average */
      idetail = icoarse + itf;
      /* UPDATE */
      vec[icoarse] 
	= vec[icoarse] - .5 * vec[idetail];
      /* PREDICT */
      vec[idetail] 
	= vec[idetail] + vec[icoarse];
    }
  }
} 
