#include "haar.h"
#include <math.h>
#include <malloc.h>
#include <stdio.h>
/*   
      Print wavelet transform intermediate results.

      The wavelet transform calculates a set of averages and
      coefficients (differences).  These are calculated recursively.
      For example, for a time series with size 32, the first iteration
      of the recursive wavelet transform will calculate a set of 16
      averages and 16 coefficients (differences).

      The next set of differences and avarages will be calculated at
      the start of the array (0..15) and will consist of 8 averages
      and 8 differences.

      This function is passed a value for 'N' which is the size
      of the (average + difference) spectrum.  It prints each of these.

*/
void haar_printintermediate(double *ts, int N) {
  const char* fmt = "%9.3e ";
  double avg;
  int i;
  int half = N / 2;
  
  avg = 0.;
  printf(" even = ");
  for (i = 0; i < half; i++) {
    printf(fmt, ts[i] );
    avg += ts[i];
  }
  printf("\n");
  printf(" odd  = ");
  for (i = half; i < N; i++) {
    printf(fmt, ts[i] );
  }
  printf("\n");
  printf("even average %20.12le\n",avg/N);
  printf("\n");
}

/*
  Predict step of the wavelet Lifting Scheme.
  
  The term "predict" comes from the articles
  'Building Your Own Wavelets at Home' from Sweldens

  - Transform

  In the wavelet transform, calculate the difference between 
  an odd element and its approximation-interpolation with 
  even coarser points. 
   
  - Inverse Transform

  Reverse the transform predict step by adding the coarse approximation
  (even element) to the difference-detail (odd element).

  - Explanation 

  Predict an odd point using Haar approximation.
  The odd point is the average of neighbouring odd and even points.
  
  The difference between the predicted (averaged) value
  and the actual odd value replaces the odd value in the
  forward transform.
  
  As the recursive steps proceed, N will eventually be 4 and
  then 2.  
  
  Parameters
  - vec       : the input data on which the forward or inverse
                transform is calculated.
  - N         : the area of vec over which the transform 
                is calculated
  - direction : (1) forward or (-1) inverse transform

*/
void haar_predict( double* vec, int N, int dir ) {
  int half = N / 2;
  int i,iphalf;
  double predictVal;

  for (i = 0; i < half; i++) {
    /*  Prediction of a value thanks to an interpolation */
    predictVal = vec[i];
    /* At index [i + half position] is the detail we
       are looking at */
    iphalf = i + half;
    /* Add or retrieve the predicted value whether
       the forward or inverse transform is applied */
    vec[iphalf] = vec[iphalf] - dir * predictVal;
  } 
} 


/*
  For the first N elements of 'ts', move the odd elements
  to the second half of the vector, leaving the even elements in the first
  half.

  In terms of the wavelet transform, this divides the vector
  into an average region and a coefficient (difference) region.

  For example, if we label the even elements with A and the 
  odd elements with B the input vector will contain

  A0 B0 A1 B1 A2 B2 A3 B3 
  
  The split function will move the odd elements into the second half
  of the array, resulting in

  A0 A1 A2 A3 B0 B1 B2 B3 

*/
void haar_split( double* ts, double* tmpvec, int N ) {
  int half = N/2;
  int i;
  
  /* Copy ts into tmpvec */
  for (i=0; i < N; i++) 
    tmpvec[i] = ts[i]; 
  /* perform the re-indexing */
  for (i=0; i < half; i++) {
    ts[i]      = tmpvec[2*i];
    ts[half+i] = tmpvec[2*i+1];
  }

}


/*
  Merge the odd and even values. This is the inverse of the
  split routine.
  
  If the even values are labeled with A and the odd values
  are labeled with B, the input vector will contain

  A0 A1 A2 A3 B0 B1 B2 B3 
  
  The merge function will re-order the odd and even element in the
  vector:

  A0 B0 A1 B1 A2 B2 A3 B3 

*/
void haar_merge( double* ts, double* tmpvec, int N ) {
  int half = N/2;
  int i;
  
  /* Copy ts into tmpvec */
  for (i=0; i < N; i++) tmpvec[i] = ts[i]; 
  /* perform the re-indexing */
  for (i=0; i < half; i++) {
    ts[2*i]      = tmpvec[i];
    ts[2*i+1]    = tmpvec[half+i];
  }
}


/*
  The update step of the wavelet Lifting Scheme
  
  In the Lifting Scheme transform the update step follows
  the predict step.  After the predict step, the differences
  have been calculated in-place writing over the even 
  values.  
*/
void haar_update( double* vec, int N, int dir ) {
  int half = N / 2;
  int i, iphalf;
  double updateVal;
  
  for (i = 0; i < half; i++) {
    /* At index [i + half position] is the detail we
       are looking at */
    iphalf = half + i;
    
    updateVal = .5 * vec[iphalf];
    
    /* Add or retrieve the update value to the even
       points whether the forward or inverse transform 
       is applied */
    vec[i] = vec[i] + dir * updateVal;
  } 
} 
    //    haar_printintermediate( ts, j);
    //    haar_printintermediate( ts, tssize);

/*
  Wavelet forward transform. version 1.
  
*/
void haar_ftransform(double *ts, 
		     double *tmpvec, int N) {
  int size;
  for (size = N; size > 1; size /= 2) {
    haar_split(ts, tmpvec, size);// rearrange ts
    haar_predict(ts, size, 1);    
    haar_update(ts, size, 1);     
  } 
} 

/*
  Wavelet inverse transform. version 1.
*/
void haar_itransform(double *ts, 
		     double *tmpvec, int N) {
  int size;
  for (size = 2; size <= N; size *= 2) {
    haar_update(ts, size, -1);    
    haar_predict(ts, size, -1);    
    haar_merge(ts, tmpvec, size);// rearrange ts
  }
}

inline double threshold_func(double threshold, double norm, int level) {
  return(threshold * pow(norm, ((double)level)));
}  
/* Thresholding of coefficients */
void haar_thresholding(double *vec, int N, 
		       double norm, 
		       double threshold) {
  register int level, i;
  /* number of non-zero coeff. at one level */
  register int nnz_level;
  int nnz_tot; /* total nb. of non-zero */
  register int half;
  register double threshold_level;
  nnz_tot = 0;
  for (level=0, half=1; half < N; 
       half *= 2, level ++) {
    threshold_level = 
      threshold_func(threshold,norm,level);
    nnz_level = 0;
    for (i = half; i < 2*half; i++) {
      if (fabs(vec[i]) < threshold_level) {
	vec[i] = 0.;
      } else {
	nnz_level++;      
      }
    }
    printf("level %4d threshold %20e nnz %10d\n",
	   level,threshold_level,nnz_level);
    nnz_tot += nnz_level;
  }
  printf("Number of non-zero coefficients :"\
	 "%13d over %13d (%.7f percents)\n",
	 nnz_tot,N,(100.*nnz_tot)/N);
}

/*
  Wavelet forward transform. version 2.
*/
void haar_fdwt(double* vec, double *ts, int N) {
  register int half, k;
  register int idetail, icoarse;
  /* loop from finest level to coarsest level */
  for (half = N/2; half >= 1; half /= 2) {
    /* Copy input 'vec' to tempory 'ts' */
    for (k = 0; k < 2*half; k++) 
      ts[k] = vec[k];

    /* loop on all coefficients at this level */
    for (k = 0; k < half; k ++) {
      /* At index 'idetail': the difference
	      at 'icoarse': the average */
      icoarse      = k;
      idetail       = half + k;
      /* PREDICT */
      vec[idetail]  = ts[2*k+1] - ts[2*k];
      /* UPDATE */ 
      vec[icoarse] = 
	ts[2*k] + .5 * vec[idetail];
    }
  }
} 

/*
  Wavelet inverse transform. version 2.
*/
void haar_idwt(double* vec, double *ts, int N) {
  register int half, i;
  register int idetail, icoarse;
  /* loop from coarsest level to finest level */
  for (half = 1; half <= N/2; half *= 2) {
    /* loop on all coefficients at this level */
    for (i = 0; i < half; i ++) {
      /* At index 'idetail' the detail will be 
	 store and at 'icoarse' is the average 
	 will be store */
      icoarse  = i;
      idetail   = half + i;
      /* UPDATE */
      ts[2*i]   = 
	vec[icoarse] - .5 * vec[idetail];
      /* PREDICT */
      ts[2*i+1] = 
	vec[idetail] + ts[2*i];
    }
    /* Copy tempory 'ts' to input 'vec' */
    for (i=0; i < 2*half; i++) vec[i] = ts[i];
  }
} 
