#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <assert.h>
#include <math.h>
#ifdef MKL
#include <mkl.h>
#else
#include <cblas.h>
#endif
#include "clock.h"
#include "matmul.h"

#ifndef TSIZE
#define TSIZE 256
#endif
#ifndef REPEAT
#define REPEAT 1
#endif

void mat_mul_basic(double*A, double*tB, 
		   double*tC, int N) {
  register double sum;
  register int i,j,k; /* iterators */
  for (i=0; i<N; i++)
    for (j=0; j<N; j++) {
      for (sum=0., k=0; k<N; k++) {
        sum += A[k+i*N]*tB[k+j*N];
      }
      tC[i+j*N] = sum;
    }
}

void mat_mul_block(double*A, double*tB, 
		   double*tC, int N, 
		   int blocki, int blockj, 
		   int blockk) {
  register int i, j, k;       
  /* block iterators */
  int          bki, bkj, bkk; 
  /* upper bounds */
  int          maxi, maxj, maxk;
  for (i=0; i<N; i++)
    for (j=0; j<N; j++) 
      tC[i+j*N] = 0.;
   
  for (bki=0; bki<N; bki+=blocki) {
    maxi=MIN(N,bki+blocki);
    for (bkj=0; bkj<N; bkj+=blockj) {
      maxj=MIN(N,bkj+blockj);
      for (bkk=0; bkk<N; bkk+=blockk) {
	maxk=MIN(N,bkk+blockk);
	for (i=bki; i!=maxi; i++)
	  for (j=bkj; j!=maxj; j++) 
	    for (k=bkk; k!=maxk; k++) 
	      tC[i+j*N] += 
		A[k+i*N]*tB[k+j*N];
      }
    }
  }
}

void mat_mul_dgemm(double*A, double*tB, 
		   double*tC, int N) {
  double alpha = 1.;
  double beta  = 0.;
#ifdef MKL
  char *notransp  = "N"; 
  char *transpos  = "T"; /* transpose */
  dgemm(transpos, notransp, &N, &N, &N, 
	&alpha, A, &N, tB, &N, &beta, 
	tC, &N);
#else
  cblas_dgemm(CblasColMajor, CblasTrans, 
	      CblasNoTrans, N, N, N, alpha, 
	      A, N, tB, N, beta, tC, N);
#endif
}



/* Perform several times the call to the subroutine
   'mat_mul_basic'. The number of times 
   is parameterized by 'repeat' parameter.
*/ 
void repeat_mat_mul_basic(double*A, double*tB, double*tC, 
			    int N, int repeat) {
  int g;
//#pragma omp barrier
  for (g=0; g<repeat; g++) {
    mat_mul_basic(A,tB,tC,N);
  }
//#pragma omp barrier
}

/* Perform several times the call to the subroutine
   'mat_mul_basic'. The number of times 
   is parameterized by 'repeat' parameter.
*/ 
void repeat_mat_mul_dgemm(double*A, double*tB, double*tC, 
			    int N, int repeat) {
  int g;
//#pragma omp barrier
  for (g=0; g<repeat; g++) {
    mat_mul_dgemm(A,tB,tC,N);
  }
//#pragma omp barrier
}

/* Perform several times the call to the subroutine
   'mat_mul_block'. The number of times 
   is parameterized by 'repeat' parameter.
*/ 
inline void repeat_mat_mul_block(double*A, double*tB, double*tC, 
			  int N, int repeat, int blocki, int blockj, 
			  int blockk) {
  int g;
//#pragma omp barrier
  for (g=0; g<repeat; g++) {
    mat_mul_block(A,tB,tC,N,blocki,blockj,blockk);
  }
//#pragma omp barrier
}

/* Benchmarks of matrix-multiply.
*/
int main(int argc, char **argv) {
  /* Number of nodes stored in a single tree */
  const int Nmax    = TSIZE;
  /* Two variables that stores the sum
     of values stored into the result matrices */
  double sumval_block = 0; 
  double sumval_dgemm = 0; 
 
//#pragma omp parallel reduction(+:sumval_block,sumval_dgemm) 
  {
    /* Two matrices we want to multiply*/
    double *A, *tB;
    /* Two matrices to store results A*B */
    double *tCbasic, *tCblock, *tCdgemm;
    /* Tmp storage */
    double *Mtmp;
    /* size of Tmp storage in order to flush caches */
    int tmpsize    = 16*1024*1024;
    /* size of matrices in Bytes*/
    int NbBytes ;
    /* Number of Bytes stored in a data structure */
    /* Number of times we perform the same algorithm
       in order to precisely estimate the execution time 
       of one call */
    int NbTimes=REPEAT;
    /* String that stores an output to print */
    char message[500];
    /* Iterators */
    int m, q, r;

    /* Alloc and fill two input matrices */
    A      = valloc(Nmax*Nmax*sizeof(double));
    tB     = valloc(Nmax*Nmax*sizeof(double));
    tCbasic = valloc(Nmax*Nmax*sizeof(double));
    tCblock = valloc(Nmax*Nmax*sizeof(double));
    tCdgemm = valloc(Nmax*Nmax*sizeof(double));
    Mtmp   = valloc(tmpsize*sizeof(double));
    for (q=0; q<Nmax; q++)
      for (r=0; r<Nmax; r++) {
	A[r+q*Nmax] = drand48();//1./(Nmax*Nmax);
	tB[q+r*Nmax] = drand48(); //(q==r?1.:0.);
	tCbasic[q+r*Nmax] = 0;
	tCblock[q+r*Nmax] = 0;
	tCdgemm[q+r*Nmax] = 0;
      }

    /* flush cache memory */
    for (q=0; q<tmpsize; q++) { Mtmp[q]=0; }

    /* Perform several times matrix multiplications.
       The total execution tqme is printed out at the end. 
    */
    NbBytes = Nmax*Nmax*sizeof(double)/1024;
    sprintf(message,"repeat_mat_mul_basic, Mem size=%d KB, time=",
	    NbBytes);
    GETTIME(repeat_mat_mul_basic(A,tB,tCbasic,Nmax,NbTimes),
	    message);

    /* flush cache memory */
    for (q=0; q<tmpsize; q++) { Mtmp[q]=0; }

    /* Perform several times matrix multiplications.
       The total execution tqme is printed out at the end. 
    */
    NbBytes = Nmax*Nmax*sizeof(double)/1024;
    sprintf(message,"repeat_mat_mul_dgemm, Mem size=%d KB, time=",
	    NbBytes);
    GETTIME(repeat_mat_mul_dgemm(A,tB,tCdgemm,Nmax,NbTimes),
	    message);

    /* flush cache memory */
    bzero(Mtmp,tmpsize*sizeof(double));
    /* Perform several times matrix multiplications.
       The total execution time is printed out at the end. 
    */
    sprintf(message,"repeat_mat_mul_block, Mem size=%d KB, time=",
	    (int)(Nmax*Nmax*sizeof(double)/1024));
    q = 128; r = 32; m= 128;
//#pragma omp barrier
    GETTIME(repeat_mat_mul_block(A,tB,tCblock,Nmax,NbTimes,q,r,m),
	    message);
    
    /* Perform parametric study to estimate block size of line 205 */ 

//    for (q=32; q<512; q*=2)
//      for (r=32; r<512; r*=2)
//	for (m=32; m<512; m*=2) {
//	  /* flush cache memory */
//	  bzero(Mtmp,tmpsize*sizeof(double));
//	  sprintf(message,"repeat_mat_mul_block, %d KB, block%d_%d_%d, time=",
//		  (int)(Nmax*Nmax*sizeof(double)/1024),q,r,m);
////#pragma omp barrier
//	  GETTIME(repeat_mat_mul_block(A,tB,tCblock,Nmax,NbTimes,q,r,m),
//		  message);
//	}
    
    for (q=0; q<Nmax; q++)
      for (r=0; r<Nmax; r++) {
	sumval_block+= fabs(tCblock[r+q*Nmax]-tCbasic[r+q*Nmax]);
	sumval_dgemm+= fabs(tCdgemm[r+q*Nmax]-tCbasic[r+q*Nmax]);
	if (fabs(tCdgemm[r+q*Nmax]-tCbasic[r+q*Nmax]) > 1e-7)
	  printf("diff %d %d = %f %f\n",r,q,tCdgemm[r+q*Nmax],tCbasic[r+q*Nmax]);
      }
    
  }
  /* Print sums obtained on the result matrices.
     This is useful to avoid compiler optimization that
     eventually suppress calls to functions that are 
     detected as unuseful. */ 
  printf("Next values should be close to zero, i.e. absolute value < 1e-4.\n");
  printf("Linf(Cbasic-Cblock) = %.16e\nLinf(Cbasic-Cdgemm) = %.16e\n",
	 sumval_block, sumval_dgemm);
  return 0;
}
